import copy
import math

from balans.base_state import _State
from balans.utils import Constants


def crossover(current: _State, rnd_state) -> _State:
    #  Take one RANDOM solutions.
    #  If a DISCRETE variable x_rand = x_inc, do not change it.
    #  Otherwise, put it to the destroy set.
    #  Send the destroy set to base_instance.
    print("*** Operator: ", "CROSSOVER")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # Static features from the instance
    discrete_indexes = current.instance.discrete_indexes

    r1_index_to_val, _ = current.instance.mip.solve_random_and_undo(Constants.timelimit_random_feasible)

    # If we don't find a random feasible solution
    if len(r1_index_to_val) == 0:
        return next_state

    #  If a discrete variable x_rand1 = x_inc, do not change it.
    indexes_with_same_value = [i for i in discrete_indexes if
                               math.isclose(r1_index_to_val[i], current.index_to_val[i])]

    # Else potentially change it
    next_state.destroy_set = set([i for i in discrete_indexes if i not in indexes_with_same_value])

    return next_state






import copy

from balans.base_state import _State


def random_objective(current: _State, rnd_state) -> _State:
    print("*** Operator: ", "RANDOM OBJECTIVE")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    next_state.has_random_obj = True

    return next_state



import copy

from balans.base_state import _State


def _mutation(current: _State, rnd_state, delta) -> _State:
    # Take the discrete variables of the original problem
    # Then, depending on delta parameter, choose a random subset of discrete variables to destroy.
    # Send the destroy set to base_instance.

    print("*** Operator: ", "MUTATION")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    # Static features from the instance
    discrete_indexes = current.instance.discrete_indexes

    destroy_size = int(delta * len(discrete_indexes))
    next_state.destroy_set = set(rnd_state.choice(discrete_indexes, destroy_size, replace=False))

    return next_state


def mutation_05(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.05)


def mutation_10(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.10)


def mutation_15(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.15)


def mutation_20(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.20)


def mutation_25(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.25)


def mutation_30(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.30)


def mutation_35(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.35)


def mutation_40(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.40)


def mutation_45(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.45)


def mutation_50(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.50)


def mutation_55(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.55)


def mutation_60(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.60)


def mutation_65(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.65)


def mutation_70(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.70)


def mutation_75(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.75)


def mutation_80(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.80)


def mutation_85(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.85)


def mutation_90(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.90)


def mutation_95(current: _State, rnd_state) -> _State:
    return _mutation(current, rnd_state, delta=0.95)



import math
import random
from typing import Tuple, Dict, Any, List

import gurobipy as gp
from gurobipy import GRB, quicksum

from balans.base_mip import _BaseMIP
from balans.utils import Constants


class _Gurobi(_BaseMIP):

    def __init__(self, instance_path: str, n_mip_jobs: int, seed: int):
        super().__init__(seed)

        # Create Gurobi model
        self.model = gp.read(instance_path)
        self.model.Params.OutputFlag = 0
        self.model.Params.Seed = self.seed
        self.model.Params.Threads = n_mip_jobs
        # self.model.Params.Presolve = 0

        # Set variables
        self.variables = self.model.getVars()

        # Set original objective and flag if we changed from max to min.
        # We always minimize
        self.org_objective_fn = self.model.getObjective()
        self.org_objective_sense = self.model.ModelSense  # 1 for minimize, -1 for maximize
        self.is_obj_sense_changed = False
        if self.org_objective_sense == -1:
            self.model.setObjective(-1 * self.org_objective_fn, GRB.MINIMIZE)
            self.is_obj_sense_changed = True

        # Set constraints, proximity z, and a flag for objective transformation
        # These are used for incremental solve and undo solve
        self.constraints = []
        self.proximity_z = None
        self.is_obj_transformed = False

        # Gurobi specific: this auxiliary var is needed for Gurobi to handle abs() constraint in dins operator.
        self.dins_abs_var = None

    def get_obj_value(self, index_to_val) -> float:
        expr = self.org_objective_fn
        obj_val = 0
        for i in range(expr.size()):
            var = expr.getVar(i)
            coeff = expr.getCoeff(i)
            obj_val += coeff * index_to_val[var.index]

        return obj_val

    def extract_indexes(self) -> Tuple[List[Any], List[Any], List[Any]]:
        # Set indexes
        discrete_indexes = []
        binary_indexes = []
        integer_indexes = []

        for var in self.variables:
            # Variable types naming different
            if self.is_discrete(var.VType):
                discrete_indexes.append(var.index)
                if self.is_binary(var.VType):
                    binary_indexes.append(var.index)
                else:
                    integer_indexes.append(var.index)

        return discrete_indexes, binary_indexes, integer_indexes

    def extract_lp(self, discrete_indexes) -> Tuple[Dict[Any, float], float, List[Any]]:
        lp_index_to_val, lp_obj_val = self.solve_lp_and_undo()
        lp_floating_discrete_indexes = [i for i in discrete_indexes if not math.isclose(lp_index_to_val[i] % 1, 0.0)]
        return lp_index_to_val, lp_obj_val, lp_floating_discrete_indexes

    def fix_vars(self, index_to_val, skip_indexes=None) -> None:
        # Thinking problems: Do we always fix all the continuous variables?
        
        # If no solution given, do nothing
        if index_to_val is None:
            return

        # Walk the variables
        for var in self.variables:
            # If a value for this variable is not given, do nothing
            if var.index not in index_to_val or index_to_val[var.index] is None:
                continue
            # if skip list given, and the variable is in there, do nothing
            if skip_indexes and var.index in skip_indexes:
                continue
                
            # Variable has a value, and it's not in the skip set, FIX
            self.constraints.append(self.model.addConstr(var == index_to_val[var.index]))

    def dins(self, index_to_val, dins_set, lp_index_to_val) -> None:
        for var in self.variables:
            index = var.index
            if index in dins_set:
                # Add bounding constraint around initial lp solution
                current_lp_diff = abs(index_to_val[index] - lp_index_to_val[index])
                # Create an auxiliary variable for the absolute value
                # Same constraint as in SCIP except that it is different to  represent the absolute value.
                # In SCIP, we use abs() function
                # In Gurobi, we split the absolute into two constraints and an additional variables.
                # Standard way to linearize the abs function
                self.dins_abs_var = self.model.addVar(lb=0.0, name=f'abs_{var.VarName}')
                self.model.update()  # Update model to include new variable
                self.constraints.append(self.model.addConstr(self.dins_abs_var >= var - lp_index_to_val[index]))
                self.constraints.append(self.model.addConstr(self.dins_abs_var >= -(var - lp_index_to_val[index])))
                self.constraints.append(self.model.addConstr(self.dins_abs_var <= current_lp_diff))
            else:
                # fix the variable
                self.constraints.append(self.model.addConstr(var == index_to_val[index]))

    def local_branching(self, index_to_val, local_branching_size, binary_indexes) -> None:
        zero_binary_vars, one_binary_vars = self.split_binary_vars(self.variables, binary_indexes, index_to_val)
        # For zero_binary_vars, we add x_i
        # For one_binary_vars, we add (1 - x_i)
        zero_expr = quicksum(zero_binary_vars)
        one_expr = quicksum(1 - var for var in one_binary_vars)
        total_expr = zero_expr + one_expr
        self.constraints.append(self.model.addConstr(total_expr <= local_branching_size))

    def proximity(self, index_to_val, obj_val, proximity_delta, binary_indexes) -> None:
        # Set the flag so we can undo
        self.is_obj_transformed = True

        zero_binary_vars, one_binary_vars = self.split_binary_vars(self.variables, binary_indexes, index_to_val)
        # if x_inc=0, new objective expression is x_inc.
        # if x_inc=1, new objective expression is 1 - x_inc.
        # Drop all other vars (when not in the expr it is set to 0 by default)
        zero_expr = quicksum(zero_binary_vars)
        one_expr = quicksum(1 - var for var in one_binary_vars)

        # add cutoff constraint depending on sense, so that next state is better quality
        # a slack variable z to prevent infeasible solution, \theta = 1
        self.proximity_z = self.model.addVar(vtype=GRB.CONTINUOUS, lb=0, name='proximity_z')
        self.model.update()  # Update model to include new variable
        self.constraints.append(self.model.addConstr(self.model.getObjective() <=
                                                     obj_val * (1 - proximity_delta) + self.proximity_z))

        # M * z is to make sure model does not use z, unless needed to avoid infeasibility
        self.model.setObjective(zero_expr + one_expr + Constants.M * self.proximity_z, GRB.MINIMIZE)

    def rens(self, index_to_val, rens_float_set, lp_index_to_val) -> None:
        for var in self.variables:
            if var.index in rens_float_set:
                # Restrict discrete vars to round up and down integer version of the lp
                # EX: If var = 3.5, the constraint is var >= 3 and var <= 4
                self.constraints.append(self.model.addConstr(var >= math.floor(lp_index_to_val[var.index])))
                self.constraints.append(self.model.addConstr(var <= math.ceil(lp_index_to_val[var.index])))
            else:
                # If not in the set, fix the var to the current state
                self.constraints.append(self.model.addConstr(var == index_to_val[var.index]))

    def random_objective(self) -> None:
        # Set the flag so we can undo
        self.is_obj_transformed = True

        # Get original objective function
        expr = self.org_objective_fn

        # Extract variables and coefficients
        vars = []
        coeffs = []

        for i in range(expr.size()):
            vars.append(expr.getVar(i))
            coeffs.append(expr.getCoeff(i))

        num_vars = len(vars)
        num_keep = max(1, int(0.25 * num_vars))  # Ensure at least one variable is kept

        # Randomly select indices to keep
        indices = list(range(num_vars))
        random.shuffle(indices)
        keep_indices = set(indices[:num_keep])

        # Build new objective function with 25% coefficients
        new_obj = gp.LinExpr()
        for i in range(num_vars):
            if i in keep_indices:
                if self.is_obj_sense_changed:
                    new_obj.addTerms(-coeffs[i], vars[i])
                else:
                    new_obj.addTerms(coeffs[i], vars[i])
            else:
                # Coefficient is zero, do not add term
                pass

        # Set the new objective function
        self.model.setObjective(new_obj, GRB.MINIMIZE)

        # Set limits
        self.model.Params.SolutionLimit = 1
        self.model.Params.Heuristics = 0

    def solve_and_undo(self, time_limit_in_sc=None, solution_limit=None) -> Tuple[Dict[Any, float], float]:
        # Set limits
        if time_limit_in_sc is not None:
            self.model.Params.TimeLimit = time_limit_in_sc
        if solution_limit is not None:
            self.model.Params.SolutionLimit = solution_limit

        # Gurobi specific: Update model after adding constraints and variables
        self.model.update()

        # Optimize
        self.model.optimize()

        # Get solution
        index_to_val, obj_val = self.get_index_to_val_and_objective()

        # Free the model
        self.model.reset(0)

        # Undo limits
        if time_limit_in_sc is not None:
            self.model.Params.TimeLimit = GRB.INFINITY
        if solution_limit is not None:
            self.model.Params.SolutionLimit = 2000000000

        # Remove constraints, and reset
        for ct in self.constraints:
            self.model.remove(ct)
        self.constraints = []

        # if dins_abs_var variable is added, remove it
        if self.dins_abs_var:
            self.model.remove(self.dins_abs_var)
            self.dins_abs_var = None

        # Reset to original objective (random objective and proximity change objective)
        if self.is_obj_transformed:
            self.is_obj_transformed = False

            # Reset back to original minimize objective
            if self.is_obj_sense_changed:
                self.model.setObjective(-1 * self.org_objective_fn, GRB.MINIMIZE)
            else:
                self.model.setObjective(self.org_objective_fn, GRB.MINIMIZE)

            # if proximity_z variable is added, remove it
            if self.proximity_z:  # if proximity_delta > 0
                self.model.remove(self.proximity_z)
                self.proximity_z = None
            else:  # if random heuristic used, reset heuristics
                self.model.Params.Heuristics = 0.05  # Reset to default value

        # Gurobi specific: Update model after removals
        self.model.update()

        # Return solution
        return index_to_val, obj_val

    def solve_random_and_undo(self, time_limit_in_sc=None) -> Tuple[Dict[Any, float], float]:
        # Set limits
        if time_limit_in_sc is not None:
            self.model.Params.TimeLimit = time_limit_in_sc

        # Set random objective
        self.random_objective()

        # Gurobi specific: Update model
        self.model.update()

        # Solve
        self.model.optimize()

        index_to_val, obj_val = self.get_index_to_val_and_objective()

        # Free the model
        self.model.reset(0)

        # Restore original objective
        self.is_obj_transformed = False

        # Reset back to original minimize objective
        if self.is_obj_sense_changed:
            self.model.setObjective(-1 * self.org_objective_fn, GRB.MINIMIZE)
        else:
            self.model.setObjective(self.org_objective_fn, GRB.MINIMIZE)

        self.model.Params.SolutionLimit = 2000000000
        self.model.Params.Heuristics = 0.05  # Reset to default
        if time_limit_in_sc is not None:
            self.model.Params.TimeLimit = GRB.INFINITY

        return index_to_val, obj_val

    def solve_lp_and_undo(self) -> Tuple[Dict[Any, float], float]:
        # Solve LP relaxation
        int_vars = []
        bin_vars = []
        variables = self.model.getVars()

        for var in variables:
            if self.is_binary(var.VType):
                var.VType = "C"
                bin_vars.append(var)
            elif self.is_discrete(var.VType):
                var.VType = "C"
                int_vars.append(var)

        # Gurobi specific: Update model after removals
        self.model.update()

        # Solve
        self.model.optimize()
        lp_index_to_val, lp_obj_val = self.get_index_to_val_and_objective()

        # Get back the original model
        self.model.reset(0)

        # Revert variable types
        for var in int_vars:
            var.VType = "I"
        for var in bin_vars:
            var.VType = "B"

        # Gurobi specific: Update model after removals
        self.model.update()

        return lp_index_to_val, lp_obj_val

    def get_index_to_val_and_objective(self) -> Tuple[Dict[Any, float], float]:
        # we check if the optimized model has solutions, feasible, and is in the solved state
        if self.model.SolCount == 0 or self.model.Status == GRB.INFEASIBLE:
            return dict(), 9999999
        else:
            index_to_val = dict([(var.index, var.X) for var in self.model.getVars()])
            obj_value = self.model.ObjVal
            return index_to_val, obj_value

    @staticmethod
    def is_discrete(var_type) -> bool:
        return var_type in ("B", "I")

    @staticmethod
    def is_binary(var_type) -> bool:
        return var_type in "B"

    @staticmethod
    def split_binary_vars(variables, binary_indexes, index_to_val) -> Tuple[List[Any], List[Any]]:
        zero_binary_vars = []
        one_binary_vars = []
        for var in variables:
            if var.index in binary_indexes:
                if math.isclose(index_to_val[var.index], 0.0):
                    zero_binary_vars.append(var)
                else:
                    one_binary_vars.append(var)
        return zero_binary_vars, one_binary_vars



import math
import random
from typing import Dict, List, Tuple, Any

import pyscipopt as scip
from pyscipopt import quicksum, Expr

from balans.base_mip import _BaseMIP
from balans.utils import Constants


class _SCIP(_BaseMIP):

    def __init__(self, instance_path: str, n_mip_jobs: int, seed: int):
        super().__init__(seed)

        # Set Scip model, variables and objective
        self.model = scip.Model()
        self.model.hideOutput()
        self.model.readProblem(instance_path)
        self.model.setParam("limits/maxorigsol", 0)
        self.model.setParam("randomization/randomseedshift", self.seed)

        # Set variables
        self.variables = self.model.getVars()

        # Set original objective and flag if we changed from max to min.
        # We always minimize
        self.org_objective_fn = self.model.getObjective()
        self.org_objective_sense = self.model.getObjectiveSense()
        self.is_obj_sense_changed = False
        if self.org_objective_sense == Constants.maximize:
            self.model.setObjective(-self.org_objective_fn)
            self.model.setMinimize()
            self.is_obj_sense_changed = True

        # Set constraints, proximity z, and a flag for objective transformation
        # These are used for incremental solve and undo solve
        self.constraints = []
        self.proximity_z = None
        self.is_obj_transformed = False

    def get_obj_value(self, index_to_val) -> float:
        obj_val = 0
        for key, item in self.org_objective_fn.terms.items():
            obj_val += item * index_to_val[key[0].getIndex()]
        return obj_val

    def extract_indexes(self) -> Tuple[List[Any], List[Any], List[Any]]:

        # Set indexes
        discrete_indexes = []
        binary_indexes = []
        integer_indexes = []

        for var in self.variables:
            if self.is_discrete(var.vtype()):
                discrete_indexes.append(var.getIndex())
                if self.is_binary(var.vtype()):
                    binary_indexes.append(var.getIndex())
                else:
                    integer_indexes.append(var.getIndex())

        return discrete_indexes, binary_indexes, integer_indexes

    def extract_lp(self, discrete_indexes) -> Tuple[Dict[Any, float], float, List[Any]]:
        lp_index_to_val, lp_obj_val = self.solve_lp_and_undo()
        lp_floating_discrete_indexes = [i for i in discrete_indexes if not lp_index_to_val[i].is_integer()]
        return lp_index_to_val, lp_obj_val, lp_floating_discrete_indexes

    def fix_vars(self, index_to_val, skip_indexes=None):
        # Thinking problems: Do we always fix all the continuous variables?

        # If no solution given, do nothing
        if index_to_val is None:
            return

        # Walk the variables
        for var in self.variables:
            # If a value for this variable is not given, do nothing
            if index_to_val[var.getIndex()] is None:
                continue
            # if skip list given, and the variable is in there, do nothing
            if skip_indexes and var.getIndex() in skip_indexes:
                continue

            # Variable has a value, and it's not in the skip set, FIX
            self.constraints.append(self.model.addCons(var == index_to_val[var.getIndex()]))

    def dins(self, index_to_val, dins_set, lp_index_to_val) -> None:
        for var in self.variables:
            if var.getIndex() in dins_set:
                # Add bounding constraint around initial lp solution
                index = var.getIndex()
                current_lp_diff = abs(index_to_val[index] - lp_index_to_val[index])
                self.constraints.append(self.model.addCons(abs(var - lp_index_to_val[index]) <= current_lp_diff))
            else:
                # fix the variable
                self.constraints.append(self.model.addCons(var == index_to_val[var.getIndex()]))

    def local_branching(self, index_to_val, local_branching_size, binary_indexes) -> None:
        # Only change a subset of the binary variables, keep others fixed. e.g.,
        zero_binary_vars, one_binary_vars = self.split_binary_vars(self.variables, binary_indexes, index_to_val)

        # if current binary var is 0, flip to 1 consumes 1 unit of budget
        # if current binary var is 1, flip to 0 consumes 1 unit of budget by (1-x)
        zero_expr = quicksum(zero_var for zero_var in zero_binary_vars)
        one_expr = quicksum(1 - one_var for one_var in one_binary_vars)
        self.constraints.append(self.model.addCons(zero_expr + one_expr <= local_branching_size))

    def proximity(self, index_to_val, obj_val, proximity_delta, binary_indexes) -> None:
        # Set the flag so we can undo
        self.is_obj_transformed = True

        zero_binary_vars, one_binary_vars = self.split_binary_vars(self.variables, binary_indexes, index_to_val)
        # if x_inc=0, new objective expression is x_inc.
        # if x_inc=1, new objective expression is 1 - x_inc.
        # Drop all other vars (when not in the expr it is set to 0 by default)
        zero_expr = quicksum(zero_var for zero_var in zero_binary_vars)
        one_expr = quicksum(1 - one_var for one_var in one_binary_vars)

        # add cutoff constraint depending on sense, so that next state is better quality
        # a slack variable z to prevent infeasible solution, \theta = 1
        self.proximity_z = self.model.addVar(vtype=Constants.continuous, lb=0)
        self.constraints.append(self.model.addCons(self.model.getObjective() <=
                                                   obj_val * (1 - proximity_delta) + self.proximity_z))

        # M * z is to make sure model does not use z, unless needed to avoid infeasibility
        self.model.setObjective(zero_expr + one_expr + Constants.M * self.proximity_z, Constants.minimize)

    def rens(self, index_to_val, rens_float_set, lp_index_to_val) -> None:
        for var in self.variables:
            if var.getIndex() in rens_float_set:
                # Restrict discrete vars to round up and down integer version of the lp
                # EX: If var = 3.5, the constraint is var >= 3 and var <= 4
                self.constraints.append(self.model.addCons(var >= math.floor(lp_index_to_val[var.getIndex()])))
                self.constraints.append(self.model.addCons(var <= math.ceil(lp_index_to_val[var.getIndex()])))
            else:
                # If not in the set, fix the var to the current state
                self.constraints.append(self.model.addCons(var == index_to_val[var.getIndex()]))

    # TODO: modify random_objective to be consistent with Gurobi version,
    #  have a delta value which control how much
    #  coeffs you want to keep, and zero out others
    # SK: may be we can parameterize these? So there can be a few different ways to randomize? and we don't loose code
    def random_objective(self) -> None:
        # Set the flag so we can undo
        self.is_obj_transformed = True

        objective = Expr()
        for var in self.variables:
            coeff = random.uniform(-1, 1)
            if coeff != 0:
                objective += coeff * var
        objective.normalize()
        self.model.setObjective(objective, Constants.minimize)

        self.model.setParam("limits/bestsol", 1)
        self.model.setHeuristics(scip.SCIP_PARAMSETTING.OFF)

    def solve_and_undo(self, time_limit_in_sc=None, solution_limit=None) -> Tuple[Dict[Any, float], float]:

        # Set limits
        if time_limit_in_sc is not None:
            self.model.setParam("limits/time", time_limit_in_sc)
        if solution_limit is not None:
            self.model.setParam("limits/bestsol", solution_limit)

        # Solve
        self.model.optimize()

        # Get solution
        index_to_val, obj_val = self.get_index_to_val_and_objective()

        # Free the model
        self.model.freeTransform()

        # Undo limits
        if time_limit_in_sc is not None:
            self.model.setParam("limits/time", 1e+20)
        if solution_limit is not None:
            self.model.setParam("limits/bestsol", -1)

        # Remove constraints, and reset
        for ct in self.constraints:
            self.model.delCons(ct)
        self.constraints = []

        # Reset to original objective (random objective and proximity change objective)
        if self.is_obj_transformed:
            self.is_obj_transformed = False

            # Reset back to original minimize objective
            if self.is_obj_sense_changed:
                self.model.setObjective(-self.org_objective_fn, Constants.minimize)
            else:
                self.model.setObjective(self.org_objective_fn, Constants.minimize)

            # if proximity_z variable is added, remove it
            if self.proximity_z:  # if proximity_delta > 0
                self.model.delVar(self.proximity_z)
                self.proximity_z = None
            else:  # if random heuristic used, reset heuristics
                self.model.setHeuristics(scip.SCIP_PARAMSETTING.DEFAULT)

        # Return solution
        return index_to_val, obj_val

    def solve_random_and_undo(self, time_limit_in_sc=None) -> Tuple[Dict[Any, float], float]:

        # Set limits
        if time_limit_in_sc is not None:
            self.model.setParam("limits/time", time_limit_in_sc)

        self.random_objective()

        # Solve
        self.model.optimize()

        r1_index_to_val, r1_obj_val = self.get_index_to_val_and_objective()

        # Get back the original model
        self.model.freeTransform()
        self.model.setParam("limits/bestsol", -1)
        self.model.setHeuristics(scip.SCIP_PARAMSETTING.DEFAULT)
        if time_limit_in_sc is not None:
            self.model.setParam("limits/time", 1e+20)

        # Reset to original objective
        self.is_obj_transformed = False

        # Reset back to original minimize objective
        if self.is_obj_sense_changed:
            self.model.setObjective(-self.org_objective_fn, Constants.minimize)
        else:
            self.model.setObjective(self.org_objective_fn, Constants.minimize)

        return r1_index_to_val, r1_obj_val

    def solve_lp_and_undo(self) -> Tuple[Dict[Any, float], float]:
        # Solve LP relaxation and save it
        int_index = []
        bin_index = []
        count = 0
        variables = self.model.getVars()
        for var in variables:
            if self.is_binary(var.vtype()):
                self.model.chgVarType(var, Constants.continuous)
                bin_index.append(count)
            elif self.is_discrete(var.vtype()):
                self.model.chgVarType(var, Constants.continuous)
                int_index.append(count)
            count += 1

        self.model.optimize()
        lp_index_to_val, lp_obj_val = self.get_index_to_val_and_objective()

        # Get back the original model
        self.model.freeTransform()
        count = 0
        for var in variables:
            if count in int_index:
                self.model.chgVarType(var, Constants.integer)
            if count in bin_index:
                self.model.chgVarType(var, Constants.binary)
            count += 1

        return lp_index_to_val, lp_obj_val

    def get_index_to_val_and_objective(self) -> Tuple[Dict[Any, float], float]:
        # we check if the optimized model has solutions, feasible, and is in the solved state
        if self.model.getNSols() == 0 or self.model.getStatus() == "infeasible" or (
                self.model.getStage() != 9 and self.model.getStage() != 10):
            return dict(), 9999999
        else:
            index_to_val = dict([(var.getIndex(), self.model.getVal(var)) for var in self.model.getVars()])
            obj_value = self.model.getObjVal()
            return index_to_val, obj_value

    @staticmethod
    def is_discrete(var_type) -> bool:
        return var_type in (Constants.binary, Constants.integer)

    @staticmethod
    def is_binary(var_type) -> bool:
        return var_type in Constants.binary

    @staticmethod
    def split_binary_vars(variables, binary_indexes, index_to_val) -> Tuple[List[Any], List[Any]]:
        zero_binary_vars = []
        one_binary_vars = []
        for var in variables:
            if var.getIndex() in binary_indexes:
                if math.isclose(index_to_val[var.getIndex()], 0.0):
                    zero_binary_vars.append(var)
                else:
                    one_binary_vars.append(var)

        return zero_binary_vars, one_binary_vars



import os
import random
from typing import List, Optional, Dict, Tuple
from typing import NamedTuple
import pickle
from multiprocessing import Pool
from functools import partial

import numpy as np
from alns.ALNS import ALNS
from alns.Result import Result
from alns.accept import LateAcceptanceHillClimbing, NonLinearGreatDeluge, AlwaysAccept
from alns.accept import MovingAverageThreshold, GreatDeluge, HillClimbing
from alns.accept import RecordToRecordTravel, SimulatedAnnealing, RandomAccept
from alns.select import AlphaUCB, MABSelector, RandomSelect, RouletteWheel, SegmentedRouletteWheel
from alns.stop import MaxIterations, MaxRuntime, NoImprovement, StoppingCriterion
from mabwiser.mab import LearningPolicy

from balans.base_instance import _Instance
from balans.base_mip import create_mip_solver
from balans.base_state import _State
from balans.destroy.crossover import crossover
from balans.destroy.dins import dins
from balans.destroy.local_branching import local_branching_05, local_branching_10, local_branching_15, \
    local_branching_20, local_branching_25, local_branching_30, local_branching_35, local_branching_40, \
    local_branching_45, local_branching_50, local_branching_55, local_branching_60, local_branching_65, \
    local_branching_70, local_branching_75, local_branching_80, local_branching_85, local_branching_90, \
    local_branching_95
from balans.destroy.mutation import mutation_05, mutation_10, mutation_15, mutation_20, mutation_25, mutation_30, \
    mutation_35, mutation_40, mutation_45, mutation_50, mutation_55, mutation_60, mutation_65, mutation_70, mutation_75, \
    mutation_80, mutation_85, mutation_90, mutation_95
from balans.destroy.proximity import proximity_005, proximity_010, proximity_015, proximity_020, proximity_025, \
    proximity_030, proximity_035, proximity_040, proximity_045, proximity_05, proximity_055, proximity_060, \
    proximity_065, \
    proximity_070, proximity_075, proximity_080, proximity_085, proximity_090, proximity_095, proximity_10
from balans.destroy.random_objective import random_objective
from balans.destroy.rens import rens_05, rens_10, rens_15, rens_20, rens_25, rens_30, rens_35, rens_40, rens_45, \
    rens_50, rens_55, rens_60, rens_65, rens_70, rens_75, rens_80, rens_85, rens_90, rens_95
from balans.destroy.rins import rins_05, rins_10, rins_15, rins_20, rins_25, rins_30, rins_35, rins_40, rins_45, \
    rins_50, rins_55, rins_60, rins_65, rins_70, rins_75, rins_80, rins_85, rins_90, rins_95
from balans.repair.repair import repair
from balans.utils import Constants, check_false, check_true, create_rng


class DestroyOperators(NamedTuple):
    Crossover = crossover

    Dins = dins

    Local_Branching_05 = local_branching_05
    Local_Branching_10 = local_branching_10
    Local_Branching_15 = local_branching_15
    Local_Branching_20 = local_branching_20
    Local_Branching_25 = local_branching_25
    Local_Branching_30 = local_branching_30
    Local_Branching_35 = local_branching_35
    Local_Branching_40 = local_branching_40
    Local_Branching_45 = local_branching_45
    Local_Branching_50 = local_branching_50
    Local_Branching_55 = local_branching_55
    Local_Branching_60 = local_branching_60
    Local_Branching_65 = local_branching_65
    Local_Branching_70 = local_branching_70
    Local_Branching_75 = local_branching_75
    Local_Branching_80 = local_branching_80
    Local_Branching_85 = local_branching_85
    Local_Branching_90 = local_branching_90
    Local_Branching_95 = local_branching_95

    Mutation_05 = mutation_05
    Mutation_10 = mutation_10
    Mutation_15 = mutation_15
    Mutation_20 = mutation_20
    Mutation_25 = mutation_25
    Mutation_30 = mutation_30
    Mutation_35 = mutation_35
    Mutation_40 = mutation_40
    Mutation_45 = mutation_45
    Mutation_50 = mutation_50
    Mutation_55 = mutation_55
    Mutation_60 = mutation_60
    Mutation_65 = mutation_65
    Mutation_70 = mutation_70
    Mutation_75 = mutation_75
    Mutation_80 = mutation_80
    Mutation_85 = mutation_85
    Mutation_90 = mutation_90
    Mutation_95 = mutation_95

    Proximity_005 = proximity_005
    Proximity_010 = proximity_010
    Proximity_015 = proximity_015
    Proximity_020 = proximity_020
    Proximity_025 = proximity_025
    Proximity_030 = proximity_030
    Proximity_035 = proximity_035
    Proximity_040 = proximity_040
    Proximity_045 = proximity_045
    Proximity_05 = proximity_05
    Proximity_055 = proximity_055
    Proximity_060 = proximity_060
    Proximity_065 = proximity_065
    Proximity_070 = proximity_070
    Proximity_075 = proximity_075
    Proximity_080 = proximity_080
    Proximity_085 = proximity_085
    Proximity_090 = proximity_090
    Proximity_095 = proximity_095
    Proximity_10 = proximity_10

    Rens_05 = rens_05
    Rens_10 = rens_10
    Rens_15 = rens_15
    Rens_20 = rens_20
    Rens_25 = rens_25
    Rens_30 = rens_30
    Rens_35 = rens_35
    Rens_40 = rens_40
    Rens_45 = rens_45
    Rens_50 = rens_50
    Rens_55 = rens_55
    Rens_60 = rens_60
    Rens_65 = rens_65
    Rens_70 = rens_70
    Rens_75 = rens_75
    Rens_80 = rens_80
    Rens_85 = rens_85
    Rens_90 = rens_90
    Rens_95 = rens_95

    Rins_05 = rins_05
    Rins_10 = rins_10
    Rins_15 = rins_15
    Rins_20 = rins_20
    Rins_25 = rins_25
    Rins_30 = rins_30
    Rins_35 = rins_35
    Rins_40 = rins_40
    Rins_45 = rins_45
    Rins_50 = rins_50
    Rins_55 = rins_55
    Rins_60 = rins_60
    Rins_65 = rins_65
    Rins_70 = rins_70
    Rins_75 = rins_75
    Rins_80 = rins_80
    Rins_85 = rins_85
    Rins_90 = rins_90
    Rins_95 = rins_95

    Random_Objective = random_objective


class RepairOperators(NamedTuple):
    Repair = repair


# Type Declarations
DestroyType = (type(DestroyOperators.Crossover),
               type(DestroyOperators.Dins),
               type(DestroyOperators.Local_Branching_05),
               type(DestroyOperators.Local_Branching_10),
               type(DestroyOperators.Local_Branching_15),
               type(DestroyOperators.Local_Branching_20),
               type(DestroyOperators.Local_Branching_25),
               type(DestroyOperators.Local_Branching_30),
               type(DestroyOperators.Local_Branching_35),
               type(DestroyOperators.Local_Branching_40),
               type(DestroyOperators.Local_Branching_45),
               type(DestroyOperators.Local_Branching_50),
               type(DestroyOperators.Local_Branching_55),
               type(DestroyOperators.Local_Branching_60),
               type(DestroyOperators.Local_Branching_65),
               type(DestroyOperators.Local_Branching_70),
               type(DestroyOperators.Local_Branching_75),
               type(DestroyOperators.Local_Branching_80),
               type(DestroyOperators.Local_Branching_85),
               type(DestroyOperators.Local_Branching_90),
               type(DestroyOperators.Local_Branching_95),
               type(DestroyOperators.Mutation_05),
               type(DestroyOperators.Mutation_10),
               type(DestroyOperators.Mutation_15),
               type(DestroyOperators.Mutation_20),
               type(DestroyOperators.Mutation_25),
               type(DestroyOperators.Mutation_30),
               type(DestroyOperators.Mutation_35),
               type(DestroyOperators.Mutation_40),
               type(DestroyOperators.Mutation_45),
               type(DestroyOperators.Mutation_50),
               type(DestroyOperators.Mutation_55),
               type(DestroyOperators.Mutation_60),
               type(DestroyOperators.Mutation_65),
               type(DestroyOperators.Mutation_70),
               type(DestroyOperators.Mutation_75),
               type(DestroyOperators.Mutation_80),
               type(DestroyOperators.Mutation_85),
               type(DestroyOperators.Mutation_90),
               type(DestroyOperators.Mutation_95),
               type(DestroyOperators.Proximity_005),
               type(DestroyOperators.Proximity_010),
               type(DestroyOperators.Proximity_015),
               type(DestroyOperators.Proximity_020),
               type(DestroyOperators.Proximity_025),
               type(DestroyOperators.Proximity_030),
               type(DestroyOperators.Proximity_035),
               type(DestroyOperators.Proximity_040),
               type(DestroyOperators.Proximity_045),
               type(DestroyOperators.Proximity_05),
               type(DestroyOperators.Proximity_055),
               type(DestroyOperators.Proximity_060),
               type(DestroyOperators.Proximity_065),
               type(DestroyOperators.Proximity_070),
               type(DestroyOperators.Proximity_075),
               type(DestroyOperators.Proximity_080),
               type(DestroyOperators.Proximity_085),
               type(DestroyOperators.Proximity_090),
               type(DestroyOperators.Proximity_095),
               type(DestroyOperators.Proximity_10),
               type(DestroyOperators.Rens_05),
               type(DestroyOperators.Rens_10),
               type(DestroyOperators.Rens_15),
               type(DestroyOperators.Rens_20),
               type(DestroyOperators.Rens_25),
               type(DestroyOperators.Rens_30),
               type(DestroyOperators.Rens_35),
               type(DestroyOperators.Rens_40),
               type(DestroyOperators.Rens_45),
               type(DestroyOperators.Rens_50),
               type(DestroyOperators.Rens_55),
               type(DestroyOperators.Rens_60),
               type(DestroyOperators.Rens_65),
               type(DestroyOperators.Rens_70),
               type(DestroyOperators.Rens_75),
               type(DestroyOperators.Rens_80),
               type(DestroyOperators.Rens_85),
               type(DestroyOperators.Rens_90),
               type(DestroyOperators.Rens_95),
               type(DestroyOperators.Rins_05),
               type(DestroyOperators.Rins_10),
               type(DestroyOperators.Rins_15),
               type(DestroyOperators.Rins_20),
               type(DestroyOperators.Rins_25),
               type(DestroyOperators.Rins_30),
               type(DestroyOperators.Rins_35),
               type(DestroyOperators.Rins_40),
               type(DestroyOperators.Rins_45),
               type(DestroyOperators.Rins_50),
               type(DestroyOperators.Rins_55),
               type(DestroyOperators.Rins_60),
               type(DestroyOperators.Rins_65),
               type(DestroyOperators.Rins_70),
               type(DestroyOperators.Rins_75),
               type(DestroyOperators.Rins_80),
               type(DestroyOperators.Rins_85),
               type(DestroyOperators.Rins_90),
               type(DestroyOperators.Rins_95),
               type(DestroyOperators.Random_Objective))

RepairType = (type(RepairOperators.Repair))

AcceptType = (MovingAverageThreshold,
              GreatDeluge,
              HillClimbing,
              LateAcceptanceHillClimbing,
              NonLinearGreatDeluge,
              AlwaysAccept,
              RecordToRecordTravel,
              SimulatedAnnealing,
              RandomAccept)

SelectorType = (AlphaUCB,
                MABSelector,
                RandomSelect,
                RouletteWheel,
                SegmentedRouletteWheel)

StopType = (MaxIterations,
            MaxRuntime,
            NoImprovement,
            StoppingCriterion)


class Balans:
    """
    High-Level Architecture:

    From the input MIP file, an Instance() is created.
    The Instance() provides:
        - Seed
        - MIP model
        - LP solution and objective
        - Indices for binary, discrete variables
        - solve(operator_settings)
        - undo_solve()

    From an Instance(), a State() is created.
    The State() provides:
        - Instance
        - Solution
        - Previous solution
        - Operator settings
        - solve_and_update()
            - calls Instance.solve(operator_settings)
    From initial state, ALNS() is created.
    ALNS iterates by calling a pair of destroy_repair on State()

    Operators takes a State()
        - Destroy operators updates States.operator_settings
        - Repair operator calls State.solve_and_update()
    """

    def __init__(self,
                 destroy_ops = None, # List
                 repair_ops = None, # List
                 selector = None, # SelectorType
                 accept = None, # AcceptType
                 stop = None, # StopType
                 seed: int = Constants.default_seed,  # The random seed
                 n_mip_jobs: int = 1,  # Number of threads for the solver
                 mip_solver: str = Constants.default_solver  # MIP solver scip/gurobi
                 ):

        # Validate arguments
        if destroy_ops is None:
            destroy_ops = [DestroyOperators.Crossover,
                           DestroyOperators.Mutation_25,
                           DestroyOperators.Mutation_50,
                           DestroyOperators.Mutation_75,
                           DestroyOperators.Local_Branching_10,
                           DestroyOperators.Local_Branching_25,
                           DestroyOperators.Local_Branching_50,
                           DestroyOperators.Proximity_005,
                           DestroyOperators.Proximity_015,
                           DestroyOperators.Proximity_030,
                           DestroyOperators.Rens_25,
                           DestroyOperators.Rens_50,
                           DestroyOperators.Rens_75,
                           DestroyOperators.Rins_25,
                           DestroyOperators.Rins_50,
                           DestroyOperators.Rins_75]
        if repair_ops is None:
            repair_ops = [RepairOperators.Repair]
        if selector is None:
            best, better, accept, reject = 1, 1, 0, 0
            selector = MABSelector(scores=[best, better, accept, reject],
                                   num_destroy=len(destroy_ops),
                                   num_repair=len(repair_ops),
                                   learning_policy=LearningPolicy.ThompsonSampling())
        if accept is None:
            accept = SimulatedAnnealing(start_temperature=20, end_temperature=1, step=0.1)
        if stop is None:
            stop = MaxIterations(10)

        self._validate_balans_args(destroy_ops, repair_ops, selector, accept, stop, seed, n_mip_jobs, mip_solver)

        # Parameters
        self.destroy_ops = destroy_ops
        self.repair_ops = repair_ops
        self.selector = selector
        self.accept = accept
        self.stop = stop
        self.seed = seed
        self.n_mip_jobs = n_mip_jobs
        self.mip_solver_str = mip_solver

        # RNG
        self._rng = create_rng(self.seed)
        self.alns_seed = self._rng.randint(0, self.seed)

        # ALNS
        self.destroy_ops = destroy_ops
        self.repair_ops = repair_ops
        self.alns = None

        # Instance and the first solution
        self._instance: Optional[_Instance] = None
        self._initial_index_to_val: Optional[Dict[int, float]] = None
        self._initial_obj_val: Optional[float] = None

    @property
    def instance(self) -> _Instance:
        return self._instance

    @property
    def initial_index_to_val(self) -> Dict[int, float]:
        return self._initial_index_to_val

    @property
    def initial_obj_val(self) -> float:
        return self._initial_obj_val

    def solve(self, instance_path, index_to_val=None) -> Result:
        """
        instance_path: the path to the MIP instance file
        index_to_val: initial (partial) solution to warm start the variables

        Returns
        -------
        Result
            ALNS result object, containing the best solution and some additional statistics.
                result.best_state.solution()
                result.best_state.objective()
        """
        self._validate_solve_args(instance_path)

        # MIP is an instance of _BaseMIP created from given mip instance
        mip = create_mip_solver(instance_path, self.seed, self.n_mip_jobs, self.mip_solver_str)

        # Create instance with mip model created from mip instance
        self._instance = _Instance(mip, self.seed)

        self._initial_index_to_val, self._initial_obj_val = self._instance.initial_solve(index_to_val=index_to_val)

        print(">>> START objective:", self._initial_obj_val)
        # print(">>> START values:", self._initial_index_to_val)

        # Initial state and solution
        initial_state = _State(self.instance, self.initial_index_to_val, self.initial_obj_val,
                               previous_index_to_val=self.initial_index_to_val)

        # Create ALNS
        self.alns = ALNS(np.random.default_rng(self.alns_seed))

        # Set ALNS operators according to MIP type, and if successful, start iterating ALNS
        if self._set_alns_operators():

            # Iterate ALNS
            result = self.alns.iterate(initial_state, self.selector, self.accept, self.stop)

            # Restore the original objective results, if max problem is reversed
            if self.instance.mip.is_obj_sense_changed:
                obj_list = []
                for obj in result.statistics.objectives:
                    obj_list.append(-obj)
                result.statistics._objectives = obj_list

            print(">>> FINISH objective:", result.best_state.objective())
        else:
            result = None

        # Result run
        return result

    @staticmethod
    def _is_local_branching(op):
        return (op == DestroyOperators.Local_Branching_05 or
                op == DestroyOperators.Local_Branching_10 or
                op == DestroyOperators.Local_Branching_15 or
                op == DestroyOperators.Local_Branching_20 or
                op == DestroyOperators.Local_Branching_25 or
                op == DestroyOperators.Local_Branching_30 or
                op == DestroyOperators.Local_Branching_35 or
                op == DestroyOperators.Local_Branching_40 or
                op == DestroyOperators.Local_Branching_45 or
                op == DestroyOperators.Local_Branching_50 or
                op == DestroyOperators.Local_Branching_55 or
                op == DestroyOperators.Local_Branching_60 or
                op == DestroyOperators.Local_Branching_65 or
                op == DestroyOperators.Local_Branching_70 or
                op == DestroyOperators.Local_Branching_75 or
                op == DestroyOperators.Local_Branching_80 or
                op == DestroyOperators.Local_Branching_85 or
                op == DestroyOperators.Local_Branching_90 or
                op == DestroyOperators.Local_Branching_95)

    @staticmethod
    def _is_proximity(op):
        return (op == DestroyOperators.Proximity_005 or
                op == DestroyOperators.Proximity_010 or
                op == DestroyOperators.Proximity_015 or
                op == DestroyOperators.Proximity_020 or
                op == DestroyOperators.Proximity_025 or
                op == DestroyOperators.Proximity_030 or
                op == DestroyOperators.Proximity_035 or
                op == DestroyOperators.Proximity_040 or
                op == DestroyOperators.Proximity_045 or
                op == DestroyOperators.Proximity_05 or
                op == DestroyOperators.Proximity_055 or
                op == DestroyOperators.Proximity_060 or
                op == DestroyOperators.Proximity_065 or
                op == DestroyOperators.Proximity_070 or
                op == DestroyOperators.Proximity_075 or
                op == DestroyOperators.Proximity_080 or
                op == DestroyOperators.Proximity_085 or
                op == DestroyOperators.Proximity_090 or
                op == DestroyOperators.Proximity_095 or
                op == DestroyOperators.Proximity_10)

    def _set_alns_operators(self):

        num_destroy_removed = 0
        # If the problem has no binary, remove Local Branching and Proximity
        if len(self.instance.binary_indexes) == 0:
            for op in self.destroy_ops:
                if self._is_local_branching(op) or self._is_proximity(op):
                    num_destroy_removed += 1
                    continue
                self.alns.add_destroy_operator(op)
        # If the problem has no integer, remove Dins
        elif len(self.instance.integer_indexes) == 0:
            for op in self.destroy_ops:
                if op == DestroyOperators.Dins:
                    num_destroy_removed += 1
                    continue
                self.alns.add_destroy_operator(op)
        else:
            for op in self.destroy_ops:
                self.alns.add_destroy_operator(op)

        for op in self.repair_ops:
            self.alns.add_repair_operator(op)

        num_remaining_destroy = self.selector.num_destroy - num_destroy_removed

        # No more operators left, return failure
        if num_remaining_destroy == 0:
            return False

        # If ops are removed, re-create bandit selector with adjusted arm counter
        if num_destroy_removed > 0:
            if isinstance(self.selector, MABSelector):
                self.selector = MABSelector(scores=self.selector.scores,
                                            num_destroy=num_remaining_destroy,
                                            num_repair=self.selector.num_repair,
                                            learning_policy=self.selector.mab.learning_policy)

        # Ops added successfully
        return True

    @staticmethod
    def _validate_balans_args(destroy_ops, repair_ops, selector, accept, stop, seed, n_mip_jobs, mip_solver):

        # Destroy Type
        for op in destroy_ops:
            check_true(isinstance(op, DestroyType), TypeError("Destroy Type mismatch." + str(op)))

        # Repair Type
        for op in repair_ops:
            check_true(isinstance(op, RepairType), TypeError("Repair Type mismatch." + str(op)))

        # Selector Type
        check_true(isinstance(selector, SelectorType), TypeError("Selector Type mismatch." + str(selector)))

        # Selector Type
        check_true(isinstance(accept, AcceptType), TypeError("Selector Type mismatch." + str(accept)))

        # Stop Type
        check_true(isinstance(stop, StopType), TypeError("Stop Type mismatch." + str(stop)))

        # Seed
        check_true(isinstance(seed, int), TypeError("The seed must be an integer." + str(seed)))

        # Parallel MIP jobs
        check_true(isinstance(n_mip_jobs, int),
                   TypeError("Number of parallel jobs must be an integer." + str(n_mip_jobs)))
        check_true(n_mip_jobs != 0, ValueError("Number of parallel jobs cannot be zero." + str(n_mip_jobs)))

        # MIP solver
        check_true(isinstance(mip_solver, str), TypeError("MIP solver backend must be a string." + str(mip_solver)))
        check_true(mip_solver in ["scip", "gurobi"],
                   ValueError("MIP solver backend must be a scip or gurobi." + str(mip_solver)))

    @staticmethod
    def _validate_solve_args(instance_path):

        check_true(isinstance(instance_path, str), TypeError("Instance path must be a string: " + str(instance_path)))
        check_false(instance_path == "", ValueError("Instance cannot be empty: " + str(instance_path)))
        check_false(instance_path is None, ValueError("Instance cannot be None: " + str(instance_path)))
        check_true(os.path.isfile(instance_path), ValueError("Instance must exist: " + str(instance_path)))


class ParBalans:
    """
    ParBalans: Run several Balans configurations in parallel.
    """

    def __init__(self,
                 n_jobs: int = 1,
                 n_mip_jobs: int = 1,
                 mip_solver: str = Constants.default_solver,
                 output_dir: str = "results/",
                 balans_generator=None):
        """
        ParBalans runs several Balans configurations in parallel.
        See class members for the possible pool of configurations used to generate random Balans configs.

        Parameters
        ----------
        n_jobs: Parallel Balans runs
        n_mip_jobs: The number of threads for the underlying mip solver, only supported by Gurobi
        mip_solver: "scip" or "gurobi"
        output_dir: Saves one file per parallel Balans run as a pickle object
                    The object is tuple with three elements: obj_of_iteration, time_of_iteration, and arm_to_reward_counts
                    There N+1 iterations, including the initial solution
                    The time is cumulative runtime when the iteration happens
                    Reward counts is the overall statistics
        balans_generator: A function that generates a Balans instance for each parallel run.
                          If none given, a default random Balans generator is used.
        """

        # Set params
        self.n_jobs = n_jobs
        self.n_mip_jobs = n_mip_jobs
        self.mip_solver = mip_solver
        self.output_dir = output_dir
        self.balans_generator = balans_generator

        # If no balans generator is given, use random balans generator by default
        if self.balans_generator is None:
            self.balans_generator = ParBalans._generate_random_balans

        # Create the results directory
        os.makedirs(self.output_dir, exist_ok=True)

    def run(self, instance_path, index_to_val=None) -> Tuple[Dict[int, float], float]:
        """
        instance_path: the path to the MIP instance file
        index_to_val: initial (partial) solution to warm start the variables

        Returns
        -------
        best_index_to_val, best_obj a tuple that contains the best solution (index_to_val) dictionary
                                    and the best objective found
        """

        # Create a dummy solver to understand objective sense
        mip = create_mip_solver(instance_path, Constants.default_seed, self.n_mip_jobs, self.mip_solver)

        # Can create other config generator function, random_config() just an example
        with Pool(processes=self.n_jobs) as pool:
            best_sol_and_obj_of_job = pool.map(partial(self._solve_instance_with_balans,
                                                       instance_path=instance_path,
                                                       index_to_val=index_to_val,
                                                       balans=self.balans_generator()),
                                               range(self.n_jobs))

        # Get the best objective value and its index base on the objective sense
        if mip.org_objective_sense == -1 or Constants.maximize:
            best_index_to_val, best_obj = max(best_sol_and_obj_of_job, key=lambda t: t[1])
        else:
            best_index_to_val, best_obj = min(best_sol_and_obj_of_job, key=lambda t: t[1])
        return best_index_to_val, best_obj

    def _solve_instance_with_balans(self, idx, instance_path, index_to_val, balans):

        if index_to_val:
            result = balans.solve(instance_path, index_to_val)
        else:
            result = balans.solve(instance_path)

        if result:
            # There N+1 iterations, including the initial solution
            # The time is cumulative runtime when the iteration happens
            # Reward counts is the overall statistics
            obj_of_iteration = result.statistics.objectives
            time_of_iteration = np.cumsum(result.statistics.runtimes)
            arm_to_reward_counts = dict(result.statistics.destroy_operator_counts)

            r = [obj_of_iteration, time_of_iteration, arm_to_reward_counts]
            result_path = os.path.join(self.output_dir, f"result_{idx}.pkl")
            with open(result_path, "wb") as fp:
                pickle.dump(r, fp)
        return result.best_state.solution(), result.best_state.objective()

    @staticmethod
    def _generate_random_balans(n_mip_jobs: int = 1, mip_solver: str = Constants.default_solver) -> Balans:

        # Pool of options
        DESTROY_CATEGORIES = {"crossover": [DestroyOperators.Crossover],
                              "mutation": [DestroyOperators.Mutation_10, DestroyOperators.Mutation_20,
                                           DestroyOperators.Mutation_30,
                                           DestroyOperators.Mutation_40, DestroyOperators.Mutation_50],
                              "local_branching": [DestroyOperators.Local_Branching_10,
                                                  DestroyOperators.Local_Branching_20,
                                                  DestroyOperators.Local_Branching_30,
                                                  DestroyOperators.Local_Branching_40,
                                                  DestroyOperators.Local_Branching_50],
                              "proximity": [DestroyOperators.Proximity_020, DestroyOperators.Proximity_040,
                                            DestroyOperators.Proximity_060,
                                            DestroyOperators.Proximity_080, DestroyOperators.Proximity_10],
                              "rens": [DestroyOperators.Rens_10, DestroyOperators.Rens_20, DestroyOperators.Rens_30,
                                       DestroyOperators.Rens_40,
                                       DestroyOperators.Rens_50],
                              "rins": [DestroyOperators.Rins_10, DestroyOperators.Rins_20, DestroyOperators.Rins_30,
                                       DestroyOperators.Rins_40,
                                       DestroyOperators.Rins_50]}
        ACCEPT_TYPE = ["HillClimbing", "SimulatedAnnealing"]
        LEARNING_POLICY = ["EpsilonGreedy", "Softmax", "ThompsonSampling"]
        REPAIR_OPERATORS = [RepairOperators.Repair]
        LP_to_REWARDS = {"binary": [[1, 1, 0, 0], [1, 1, 1, 0]],
                         "numeric": [[3, 2, 1, 0], [5, 2, 1, 0], [5, 4, 2, 0], [8, 3, 1, 0],
                                     [8, 4, 2, 1], [16, 4, 2, 1]]}

        # Destroy
        num_destroy = random.randint(len(DESTROY_CATEGORIES) - 2, len(DESTROY_CATEGORIES) * 3)
        chosen_destroy_ops = []
        if num_destroy > len(DESTROY_CATEGORIES) - 1:
            # Choose at least one member from each category
            for category in DESTROY_CATEGORIES:
                element = random.choice(DESTROY_CATEGORIES[category])
                chosen_destroy_ops.append(element)

            # Remove the already chosen elements from the pool
            all_elements = [item for sublist in DESTROY_CATEGORIES.values() for item in sublist]
            remaining_pool = list(set(all_elements) - set(chosen_destroy_ops))

            # Choose the remaining elements randomly from the remaining pool
            remaining_elements = num_destroy - len(DESTROY_CATEGORIES)
            chosen_destroy_ops.extend(random.sample(remaining_pool, remaining_elements))
        else:
            # Choose the categories to be included
            chosen_categories = random.sample(list(DESTROY_CATEGORIES.keys()), num_destroy)

            # Choose one element from each chosen category
            for category in chosen_categories:
                element = random.choice(DESTROY_CATEGORIES[category])
                chosen_destroy_ops.append(element)

        # Accept
        chosen_accept_type = []
        for op in ACCEPT_TYPE:
            if "HillClimbing" in op:
                chosen_accept_type.append(HillClimbing())
            if "SimulatedAnnealing" in op:
                chosen_accept_type.append(SimulatedAnnealing(start_temperature=10,
                                                             end_temperature=1,
                                                             step=random.uniform(0.01, 1),
                                                             method="linear"))
        acceptance_obj = random.choice(chosen_accept_type)

        # Learning Policy
        chosen_learning_policy = []
        for op in LEARNING_POLICY:
            if "EpsilonGreedy" in op:
                chosen_learning_policy.append(LearningPolicy.EpsilonGreedy(epsilon=random.uniform(0.01, 0.5)))
            if "Softmax" in op:
                chosen_learning_policy.append(LearningPolicy.Softmax(tau=random.uniform(1, 3)))
            if "ThompsonSampling" in op:
                chosen_learning_policy.append(LearningPolicy.ThompsonSampling())
        chosen_lp = random.choice(chosen_learning_policy)

        # Rewards
        chosen_scores = random.choice(LP_to_REWARDS["numeric"])
        if isinstance(chosen_lp, LearningPolicy.ThompsonSampling):
            chosen_scores = random.choice(LP_to_REWARDS["binary"])

        # Seed
        chosen_seed = random.randint(1, 100000)

        # Stop
        stop = MaxIterations(10) # MaxRuntime(100)

        # Balans
        balans = Balans(destroy_ops=chosen_destroy_ops,
                        repair_ops=REPAIR_OPERATORS,
                        selector=MABSelector(scores=chosen_scores,
                                             num_destroy=len(chosen_destroy_ops),
                                             num_repair=len(REPAIR_OPERATORS),
                                             learning_policy=chosen_lp,
                                             seed=chosen_seed),
                        accept=acceptance_obj,
                        stop=stop,
                        n_mip_jobs=n_mip_jobs,
                        mip_solver=mip_solver)

        return balans



from balans.base_state import _State


def repair(current: _State, rnd_state) -> _State:
    print("\t Repair")

    # Solve the state with fixed variables to repair and update solution and objective
    current.solve_and_update()

    print("\t Repair objective:", current.obj_val)

    return current



from typing import Tuple, Dict, Any

from balans.base_mip import _BaseMIP
from balans.utils import Constants


class _Instance:
    """
    Instance from a given MIP file with solve operations on top, subject to operator
    """

    def __init__(self, mip: _BaseMIP, seed=Constants.default_seed):
        # MIP Model initialized from the original mip instance
        self.mip: _BaseMIP = mip
        self.seed = seed

        # Static, set once and for all
        self.discrete_indexes = None    # all discrete: binary + integer
        self.binary_indexes = None      # discrete and binary
        self.integer_indexes = None     # discrete but not binary
        self.lp_index_to_val = None
        self.lp_obj_val = None
        self.lp_floating_discrete_indexes = None

    def initial_solve(self, index_to_val=None) -> Tuple[Dict[Any, float], float]:

        # Extracts static instance features (sense must be minimization now, after adjustment)
        self.discrete_indexes, self.binary_indexes, self.integer_indexes = self.mip.extract_indexes()

        # Extract static lp features for root instance
        self.lp_index_to_val, self.lp_obj_val, self.lp_floating_discrete_indexes = self.mip.extract_lp(self.discrete_indexes)

        # If a solution is given fix it. It can be partial (denoted by None value). Appends mip.constraints
        self.mip.fix_vars(index_to_val)

        # Solve with some time limit to get an initial solution
        index_to_val, obj_val = self.mip.solve_and_undo(time_limit_in_sc=Constants.timelimit_first_solution)

        # If no feasible initial solution found within time limit
        if len(index_to_val) == 0:
            # Solve for the first feasible solution without time limit and without fixing the given solution
            index_to_val, obj_val = self.mip.solve_and_undo(solution_limit=1)

        # Return solution
        return index_to_val, obj_val

    def solve(self,
              index_to_val=None,
              obj_val=None,
              destroy_set=None,
              dins_set=None,
              rens_float_set=None,
              has_random_obj=False,
              local_branching_size=0,
              proximity_delta=0) -> Tuple[Dict[Any, float], float]:

        print("\t Solve")

        # has_destroy to identify if any constraint added or objective function changed
        # If has_destroy = True, optimize the problem and get new sol and obj
        # If has_destroy = False, return the current sol and obj, do not optimize
        has_destroy = False

        # Starting solution and objective value
        starting_index_to_val = index_to_val
        starting_obj_val = obj_val

        # DESTROY used for Crossover, Mutation, RINS
        if destroy_set:
            if len(destroy_set) > 0:
                has_destroy = True
                self.mip.fix_vars(index_to_val, skip_indexes=destroy_set)

        # DINS: Discrete Variables, where incumbent and lp relaxation have distance more than 0.5
        if dins_set:
            if len(dins_set) > 0:
                has_destroy = True
                self.mip.dins(index_to_val, dins_set, self.lp_index_to_val)

        # Local Branching: Binary variables, flip a limited subset (can come from DINS with delta)
        if local_branching_size > 0:
            has_destroy = True
            self.mip.local_branching(index_to_val, local_branching_size, self.binary_indexes)

        # Proximity: Binary variables, modify objective, add new constraint
        if proximity_delta > 0:
            has_destroy = True
            print("proximity_delta: ", proximity_delta)
            self.mip.proximity(index_to_val, obj_val, proximity_delta, self.binary_indexes)

        # RENS: Discrete variables, where the lp relaxation is not integral
        if rens_float_set:
            if len(rens_float_set) > 0:
                has_destroy = True
                self.mip.rens(index_to_val, rens_float_set, self.lp_index_to_val)

        # Random Objective
        if has_random_obj:
            has_destroy = True
            self.mip.random_objective()

        # If no destroy, don't solve, quit with previous objective
        # e.g. when destroy set is empty.
        if not has_destroy:
            print("No destroy to apply, don't call optimize()")
            print("\t Current Obj:", starting_obj_val)
            # print("\t starting_index_to_val: ", starting_index_to_val)
            return starting_index_to_val, starting_obj_val

        # Solve mip and undo with some timelimit
        # Set time limit per alns iteration or overwrite with local branching iteration
        time_limit = Constants.timelimit_alns_iteration
        if local_branching_size:
            time_limit = Constants.timelimit_local_branching_iteration
        index_to_val, obj_val = self.mip.solve_and_undo(time_limit_in_sc=time_limit)

        # If no solution found, go back
        if len(index_to_val) == 0:
            print("No solution found, go back to previous state")
            # print("\t Current Obj:", starting_obj_val)
            return starting_index_to_val, starting_obj_val

        # Solution found but for transformed objectives (random_obj and proximity), find the original obj value
        if proximity_delta > 0 or has_random_obj:
            # Objective value of the solution found in transformed
            print("\t Transformed obj: ", obj_val)
            obj_val = self.mip.get_obj_value(index_to_val)

        print("\t Solve DONE!", obj_val)
        return index_to_val, obj_val



import abc
from typing import Dict, List, Tuple, Any

from balans.utils import Constants


class _BaseMIP(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self, seed: int):
        self.seed = seed

        # Model, variables, objective
        self.model = None
        self.variables = None
        self.org_objective_fn = None
        self.org_objective_sense = None
        self.is_obj_sense_changed = False       # we always minimize

        # These are used for incremental solving
        self.constraints = []
        self.proximity_z = None          # Used in proximity
        self.is_obj_transformed = False  # Used in proximity and random objective

    @abc.abstractmethod
    def get_obj_value(self, index_to_val) -> float:
        """ Given a solution, return objective value """
        pass

    @abc.abstractmethod
    def extract_indexes(self) -> Tuple[List[Any], List[Any], List[Any]]:
        """ Return discrete, binary, and integer index lists"""
        pass

    @abc.abstractmethod
    def extract_lp(self, discrete_indexes) -> Tuple[Dict[Any, float], float, List[Any]]:
        """ Return lp index_to_val, lp objective value, and the floating discrete indexes """
        pass

    @abc.abstractmethod
    def fix_vars(self, index_to_val, skip_indexes=None) -> None:
        """ Add constraints to fix variables to given values except skipped vars"""
        pass

    @abc.abstractmethod
    def dins(self, index_to_val, dins_set, lp_index_to_val) -> None:
        pass

    @abc.abstractmethod
    def local_branching(self, index_to_val, local_branching_size, binary_indexes) -> None:
        pass

    @abc.abstractmethod
    def proximity(self, index_to_val, obj_val, proximity_delta, binary_indexes) -> None:
        pass

    @abc.abstractmethod
    def rens(self, index_to_val, rens_float_set, lp_index_to_val) -> None:
        pass

    @abc.abstractmethod
    def random_objective(self) -> None:
        pass

    @abc.abstractmethod
    def solve_and_undo(self, time_limit_in_sc=None, solution_limit=None) -> Tuple[Dict[Any, float], float]:
        """
        Solve with the given time and solution limit, return the solution index_to_val and obj value
        Make sure to undo the solve and clear the constraints, promixity z, and reset objective.
        """
        pass

    @abc.abstractmethod
    def solve_random_and_undo(self, time_limit_in_sc=None) -> Tuple[Dict[Any, float], float]:
        """
        Solve with the given time limit and return a random solution index_to_val and obj value
        Make sure to undo the solve operation.
        """
        pass

    @abc.abstractmethod
    def solve_lp_and_undo(self) -> Tuple[Dict[Any, float], float]:
        """
        Solve lp relaxation and undo.
        """
        pass


def create_mip_solver(instance_path: str,
                      seed: int = Constants.default_seed,
                      n_mip_jobs: int = 1,
                      mip_solver: str = Constants.default_solver) -> _BaseMIP:
    """ Returns a mip model of the given solver type for the given instance

        Parameters
        ----------
        instance_path: str
            the path to mip instance
        seed: int
            the seed to pass to mip model
        mip_solver : string
            the type of the mip model, scip or gurobi

        Returns
        -------
        out : _BaseMIP
            A MIP Solver object that implements the base solver class
    """
    from balans.solver_gurobi import _Gurobi
    from balans.solver_scip import _SCIP

    mip_factory = {Constants.gurobi_solver: _Gurobi,
                   Constants.scip_solver: _SCIP,}

    return mip_factory.get(mip_solver)(instance_path, n_mip_jobs, seed)



import os

import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    required = f.read().splitlines()

with open(os.path.join('balans', '_version.py')) as f:
    exec(f.read())

# python setup.py bdist_wheel
setuptools.setup(
    name="balans",
    description="Balans: Bandit-based Adaptive Large Neighborhood Search",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=__version__,
    author="skadio",
    url="https://github.com/skadio/balns",
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "notebooks"]),
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentation": "https://github.com/skadio/balans",
        "Source": "https://github.com/skadio/balans"
    }
)



import copy

from balans.base_state import _State


def proximity(current: _State, rnd_state, delta=0.005) -> _State:
    # Objective function modification
    # Change the objective coefficients of the original
    # problem based on the current solution value.
    # For binary variables,
    # if x_inc =0, update its objective coefficient to 1.
    # if x_inc =1, update its objective coefficient to -1.
    # Drop all other variables by making their coefficient 0.
    # Send the destroy set to base_instance.
    # Note : the required objective operations for proximity search happens in base_instance file.

    print("*** Operator: ", "PROXIMITY SEARCH")
    print("\t Destroy current objective:", current.obj_val)
    next_state = copy.deepcopy(current)
    next_state.reset_solve_settings()

    next_state.proximity_delta = delta

    return next_state


def proximity_005(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.005)


def proximity_010(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.010)


def proximity_015(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.015)


def proximity_020(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.020)


def proximity_025(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.025)


def proximity_030(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.030)


def proximity_035(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.035)


def proximity_040(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.040)


def proximity_045(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.045)


def proximity_05(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.05)


def proximity_055(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.055)


def proximity_060(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.060)


def proximity_065(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.065)


def proximity_070(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.070)


def proximity_075(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.075)


def proximity_080(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.080)


def proximity_085(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.085)


def proximity_090(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.090)


def proximity_095(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.095)


def proximity_10(current: _State, rnd_state) -> _State:
    return proximity(current, rnd_state, delta=0.10)
